;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "bicycle" "1.1.3"
  "Cycle outline and code visibility."
  '((emacs  "28.1")
    (compat "30.1"))
  :url "https://github.com/tarsius/bicycle"
  :commit "b8919e56e7f20e4164002e9b4dba0268779b5645"
  :revdesc "v1.1.3-0-gb8919e56e7f2"
  :keywords '("outlines")
  :authors '(("Jonas Bernoulli" . "emacs.bicycle@jonas.bernoulli.dev"))
  :maintainers '(("Jonas Bernoulli" . "emacs.bicycle@jonas.bernoulli.dev")))
